package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.model.core.AbstractSpace;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;

import java.util.ArrayList;
import java.util.List;

public class TrafficSpaceObject extends TransportationObject {

    private final List<TrafficAreaObject> trafficAreas = new ArrayList<>();

    public enum TrafficSpaceType {
        TRAFFIC_SPACE, AUXILIARY_TRAFFIC_SPACE
    }

    public TrafficSpaceObject(TrafficSpaceType trafficSpaceType) {
        if (trafficSpaceType == TrafficSpaceType.TRAFFIC_SPACE) {
            super.setType(TransportationType.TRAFFIC_SPACE);
        } else {
            super.setType(TransportationType.AUXILLIARY_TRAFFIC_SPACE);
        }
    }

    public List<TrafficAreaObject> getTrafficAreas() {
        return trafficAreas;
    }

    public void addTrafficArea(TrafficAreaObject trafficAreaObject) {
        trafficAreas.add(trafficAreaObject);
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
            AbstractSpace ats = (AbstractSpace) super.getGmlObject();
            setMultiSurfaceAccordingToLod(ats, ms, geom.getLod());
        }
        for (TrafficAreaObject tao : trafficAreas) {
            tao.reCreateGeometries(factory, config);
        }
    }

    private void setMultiSurfaceAccordingToLod(AbstractSpace ats, MultiSurface ms, Lod lod) {
        switch (lod) {
            case LOD0:
                ats.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                ats.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                ats.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("cannot set geometry with LOD for AuxiliaryTrafficSpace: " + lod);
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (TrafficAreaObject tao : trafficAreas) {
            if (tao.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (TrafficAreaObject tao : trafficAreas) {
            tao.clearAllContainedCheckResults();
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (TrafficAreaObject tao : trafficAreas) {
            tao.collectContainedErrors(errors);
        }
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (TrafficAreaObject tao : trafficAreas) {
            if (tao.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (TrafficAreaObject tao : trafficAreas) {
            tao.accept(c);
        }
    }

    @Override
    public void unsetGmlGeometries() {
        AbstractSpace ats = (AbstractSpace) super.getGmlObject();
        ats.setLod0MultiSurface(null);
        ats.setLod2MultiSurface(null);
        ats.setLod3MultiSurface(null);
        for (TrafficAreaObject tao : trafficAreas) {
            tao.unsetGmlGeometries();
        }
    }

    @Override
    public CityObject getTopLevelCityObject() {
        return this;
    }


    @Override
    public String toString() {
        return "TransportationSpaceObject [id=" + getGmlId() + "]";
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (TrafficAreaObject tao : trafficAreas) {
            tao.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (TrafficAreaObject tao : trafficAreas) {
            tao.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        for (TrafficAreaObject tao : trafficAreas) {
            handler.addInstance(tao);
        }
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        TrafficSpaceObject originalTao = (TrafficSpaceObject) original;
        for (TrafficAreaObject tao : originalTao.trafficAreas) {
            trafficAreas.add(handler.getCopyInstance(tao));
        }
        super.setGmlObject(originalTao.getGmlObject());
    }

    @Override
    public Copyable createCopyInstance() {
        return new TrafficSpaceObject(null);
    }


}
